#!/usr/bin/env python3

import click
import json
import os
import requests
import gzip
import urllib
import shutil
import logging
import traceback
from ontobio.model.association import Curie, ExtensionUnit
from ontobio.io.entityparser import GpiParser
from ontobio.ontol_factory import OntologyFactory
from ontobio.io.gafparser import GafParser
from ontobio.io.gpadparser import GpadParser
from ontobio.io.assocwriter import GafWriter
from ontobio.io.assocwriter import GpadWriter
from ontobio.io import assocparser
from ontobio.io import gafgpibridge
from ontobio.io import entitywriter
from ontobio.io import gaference
from ontobio.rdfgen import assoc_rdfgen
from ontobio.rdfgen.gocamgen.gocam_builder import GoCamBuilder, AssocExtractor
from ontobio.validation import metadata
from ontobio.validation import tools
from ontobio.validation import rules

from typing import Dict, Set

# logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s: %(message)s", level=logging.WARNING)

logger = logging.getLogger("ontobio")


def thispath():
    os.path.normpath(os.path.abspath(__file__))


def download_a_dataset_source(group, dataset_metadata, target_dir, source_url, base_download_url=None,
                              replace_existing_files=True):
    """
    This will download a dataset source given the group name,
    the metadata stanza for the dataset, and the target directory that all downloads
    occur in.

    The path will be built from these elements and then the URL will be found
    from which to download the file.

    `base_download_url` if set will change the URL to a local relative path, rather
    than a true download. This means the dataset_metadata could have "relative/path/to/gaf"
    and the absolute on disk path will be constructed by appending the found metadata
    path to `base_download_url`.

    `replace_existing_files` by default is True. This will overwrite any existing file, always
    updating. With `replace_existing_files` False the path will be checked if a file already
    exists there and if so the actual download will not proceed. The found file
    will be assumed to be the correct file.
    """
    # Local target download path setup - path and then directories
    file_name = source_url.split("/")[-1]
    path = metadata.source_path(dataset_metadata, target_dir, group)

    # Just return the path if we find that it exists already
    if os.path.exists(path) and not replace_existing_files:
        click.echo("{} already exists, no need to download - skipping".format(path))
        return path

    os.makedirs(os.path.split(path)[0], exist_ok=True)

    click.echo("Downloading source to {}".format(path))

    # Parse urls
    source_url_parsed = urllib.parse.urlparse(source_url)
    base_download_url_parsed = urllib.parse.urlparse(base_download_url) if base_download_url != None else None

    # Compute scheme:
    scheme = "file"
    if source_url_parsed.scheme != "":
        scheme = source_url_parsed.scheme
    elif base_download_url_parsed.scheme != "":
        scheme = base_download_url_parsed.scheme

    # Compute joined url, including dealing with relative paths
    joined_url = source_url
    if not source_url_parsed.path.startswith("/"):
        # We're relative
        if base_download_url is not None:
            joined_url = urllib.parse.urljoin(base_download_url, source_url)
        else:
            # This is bad and we have to jump out`since we have no way of constructing a real gaf url
            raise click.ClickException(
                "Option `--base-download-url` was not specified and the config url {} is a relative path.".format(
                    source_url))

    # including scheme and such
    reconstructed_url = urllib.parse.urlunsplit((scheme, urllib.parse.urlparse(joined_url).netloc,
                                                 urllib.parse.urlparse(joined_url).path, source_url_parsed.query, ""))

    click.echo("Using URL `{}`".format(reconstructed_url))

    # Using urllib to download if scheme is ftp or file. Otherwise we can use requests and use a progressbar
    if scheme in ["ftp", "file"]:
        urllib.request.urlretrieve(reconstructed_url, path)
    else:
        response = requests.get(reconstructed_url, stream=True)
        content_length = int(response.headers.get("Content-Length", 0))

        with open(path, "wb") as downloaded:
            with click.progressbar(iterable=response.iter_content(chunk_size=512 * 1024), length=content_length,
                                   show_percent=True) as chunks:
                for chunk in chunks:
                    if chunk:
                        downloaded.write(chunk)

    return path


def download_source_gafs(group_metadata, target_dir, exclusions=[], base_download_url=None, replace_existing_files=True,
                         only_dataset=None):
    """
    This looks at a group metadata dictionary and downloads each GAF source that is not in the exclusions list.
    For each downloaded file, keep track of the path of the file. If the file is zipped, it will unzip it here.
    This function returns a list of tuples of the dataset dictionary mapped to the downloaded source path.
    """
    # Grab all datasets in a group, excluding non-gaf, datasets that are explicitely excluded from an option, and excluding datasets with the `exclude key` set to true
    gaf_urls = []
    if only_dataset is None:
        gaf_urls = [(data, data["source"]) for data in group_metadata["datasets"] if
                    data["type"] == "gaf" and data["dataset"] not in exclusions and not data.get("exclude", False)]
    else:
        gaf_urls = [(data, data["source"]) for data in group_metadata["datasets"] if data["dataset"] == only_dataset]
    # List of dataset metadata to gaf download url

    click.echo("Found {}".format(", ".join([kv[0]["dataset"] for kv in gaf_urls])))
    downloaded_paths = []
    for dataset_metadata, gaf_url in gaf_urls:
        dataset = dataset_metadata["dataset"]
        # Local target download path setup - path and then directories
        path = download_a_dataset_source(group_metadata["id"], dataset_metadata, target_dir, gaf_url,
                                         base_download_url=base_download_url,
                                         replace_existing_files=replace_existing_files)

        if dataset_metadata.get("compression", None) == "gzip":
            # Unzip any downloaded file that has gzip, strip of the gzip extension
            path = unzip_simple(path)
        else:
            # otherwise file is coming in uncompressed. But we want to make sure
            # to zip up the original source also
            tools.zipup(path)

        downloaded_paths.append((dataset_metadata, path))

    return downloaded_paths


def check_and_download_mixin_source(mixin_metadata, group_id, dataset, target_dir, base_download_url=None,
                                    replace_existing_files=True):
    mixin_dataset = tools.find(mixin_metadata["datasets"], lambda d: d.get("merges_into", "") == dataset)
    if mixin_dataset is None:
        return None

    click.echo("Downloading dataset {}".format(mixin_dataset["source"]))
    path = download_a_dataset_source(group_id, mixin_dataset, target_dir, mixin_dataset["source"],
                                     base_download_url=base_download_url, replace_existing_files=replace_existing_files)

    unzipped = unzip_simple(path)  # Strip off the .gz extension, leaving just the unzipped filename
    return unzipped


def mixin_dataset(mixin_metadata, dataset):
    mixin_dataset_version = tools.find(mixin_metadata["datasets"], lambda d: d.get("merges_into", "") == dataset)
    if mixin_dataset_version is None:
        return None  # Blah

    return mixin_dataset_version


def unzip(path, target):
    click.echo("Unzipping {}".format(path))

    def chunk_gen():
        with gzip.open(path, "rb") as p:
            while True:
                chunk = p.read(size=512 * 1024)
                if not chunk:
                    break
                yield chunk

    with open(target, "wb") as tf:
        with click.progressbar(iterable=chunk_gen()) as chunks:
            for chunk in chunks:
                tf.write(chunk)


def unzip_simple(zipped_path):
    # 'Simple' meaning no chunking like in unzip()
    unzipped = os.path.splitext(zipped_path)[0]  # Strip off the .gz extension, leaving just the unzipped filename
    unzip(zipped_path, unzipped)
    return unzipped


def create_parser(config, group, dataset, format="gaf"):
    if format == "gpad":
        return GpadParser(config=config, group=group, dataset=dataset)
    else:
        # We assume it's gaf as we only support in this instant gaf and gpad
        return GafParser(config=config, group=group, dataset=dataset)


"""
Produce validated gaf using the gaf parser/
"""


@tools.gzips
def produce_gaf(dataset, source_gaf, ontology_graph, gpipaths=None, paint=False, group="unknown", rule_metadata=None,
                goref_metadata=None, ref_species_metadata=None, db_type_name_regex_id_syntax=None,
                retracted_pub_set=None, db_entities=None, group_idspace=None,
                format="gaf", suppress_rule_reporting_tags=[], annotation_inferences=None, group_metadata=None,
                extensions_constraints=None, rule_contexts=[], gaf_output_version="2.2",
                rule_set=assocparser.RuleSet.ALL) -> list[str]:
    filtered_associations = open(os.path.join(os.path.split(source_gaf)[0], "{}_noiea.gaf".format(dataset)), "w")
    config = assocparser.AssocParserConfig(
        ontology=ontology_graph,
        filter_out_evidence=["IEA"],
        filtered_evidence_file=filtered_associations,
        gpi_authority_path=gpipaths,
        paint=paint,
        rule_metadata=rule_metadata,
        goref_metadata=goref_metadata,
        ref_species_metadata=ref_species_metadata,
        db_type_name_regex_id_syntax=db_type_name_regex_id_syntax,
        retracted_pub_set=retracted_pub_set,
        entity_idspaces=db_entities,
        group_idspace=group_idspace,
        suppress_rule_reporting_tags=suppress_rule_reporting_tags,
        annotation_inferences=annotation_inferences,
        group_metadata=group_metadata,
        extensions_constraints=extensions_constraints,
        rule_contexts=rule_contexts,
        rule_set=rule_set,
    )
    click.echo("Producing {}".format(source_gaf))
    # logger.info("AssocParserConfig used: {}".format(config))
    split_source = os.path.split(source_gaf)[0]
    validated_gaf_path = os.path.join(split_source, "{}_valid.gaf".format(dataset))
    outfile = open(validated_gaf_path, "w")
    gafwriter = GafWriter(file=outfile, source=dataset, version=gaf_output_version)

    click.echo("Validating source {}: {}".format(format, source_gaf))
    parser = create_parser(config, group, dataset, format)
    with open(source_gaf) as sg:
        lines = sum(1 for line in sg)

    with open(source_gaf) as gaf:
        with click.progressbar(iterable=parser.association_generator(file=gaf), length=lines) as associations:
            for assoc in associations:
                gafwriter.write_assoc(assoc)

    outfile.close()
    filtered_associations.close()

    report_markdown_path = os.path.join(os.path.split(source_gaf)[0], "{}.report.md".format(dataset))
    logger.info("About to write markdown report to {}".format(report_markdown_path))
    with open(report_markdown_path, "w") as report_md:
        logger.info("Opened for writing {}".format(report_markdown_path))
        report_md.write(parser.report.to_markdown())

    logger.info("markdown {} written out".format(report_markdown_path))
    logger.info("Markdown current stack:")
    if logger.getEffectiveLevel() == logging.INFO:
        traceback.print_stack()

    report_json_path = os.path.join(os.path.split(source_gaf)[0], "{}.report.json".format(dataset))
    logger.info("About to write json report to {}".format(report_json_path))
    with open(report_json_path, "w") as report_json:
        logger.info("Opened for writing {}".format(report_json_path))
        report_json.write(json.dumps(parser.report.to_report_json(), indent=4))

    logger.info("json {} written out".format(report_markdown_path))
    logger.info("gorule-13 first 10 messages: {}".format(
        json.dumps(parser.report.to_report_json()["messages"].get("gorule-0000013", [])[:10], indent=4)))
    logger.info("json current Stack:")
    if logger.getEffectiveLevel() == logging.INFO:
        traceback.print_stack()

    return [validated_gaf_path, filtered_associations.name]


@tools.gzips
def make_ttls(dataset, gaf_path, products, ontology_graph):
    with open(gaf_path) as sg:
        lines = sum(1 for line in sg)

    product_files = {
        "ttl": open(os.path.join(os.path.split(gaf_path)[0], "{}_cam.ttl".format(dataset)), "wb")
    }

    if not products["ttl"]:
        # Bail if we have no products
        return []

    # def write_gpi_entity(association, bridge, gpiwriter):
    with open(gaf_path) as gf:
        gafparser = GafParser(config=assocparser.AssocParserConfig(
            ontology=ontology_graph,
            paint=True,
        ))

        # gpi info:
        click.echo("Using {} as the gaf to build data products with".format(gaf_path))
        if products["ttl"]:
            click.echo("Setting up {}".format(product_files["ttl"].name))
            rdf_writer = assoc_rdfgen.TurtleRdfWriter(label=os.path.split(product_files["ttl"].name)[1])
            transformer = assoc_rdfgen.CamRdfTransform(writer=rdf_writer)

        click.echo("Making ttl products...")
        with click.progressbar(iterable=gafparser.association_generator(file=gf), length=lines) as associations:
            for association in associations:
                if products["ttl"]:
                    transformer.provenance()
                    transformer.translate(association)

        # post ttl steps
        if products["ttl"]:
            click.echo("Writing ttl to disk")
            rdf_writer.serialize(destination=product_files["ttl"])

        # After we run through associations
        for f in product_files.values():
            f.close()

    return [product_files[prod].name for prod in sorted(product_files.keys()) if products[prod]]


@tools.gzips
def make_gpads(dataset, gaf_path, products, ontology_graph,
               noctua_gpad_file, paint_gaf_src, gpi, gpad_gpi_output_version):
    """
    Using the gaf files and the noctua gpad file, produce a gpad file that contains both kinds of annotations
    without any loss.

    :param dataset: The dataset name
    :param gaf_path: The path to the gaf file
    :param products: The products to make
    :param ontology_graph: The ontology graph to use for parsing the associations
    :param noctua_gpad_file: The path to the noctua gpad file
    :param paint_gaf_src: The source of the paint gaf file
    :param gpi: The path to the gpi file -- needed to convert isoform annotations from Noctua files
                                            to gene annotations in GAF outputs.
    :return: The path to the gpad file

    """
    gpad_file_path = os.path.join(os.path.split(gaf_path)[0], f"{dataset}.gpad")

    if not products["gpad"]:
        return []

    # Open the file once and keep it open for all operations within this block
    with open(gpad_file_path, "w") as outfile:
        gpadwriter = GpadWriter(file=outfile, version=gpad_gpi_output_version)

        # If there's a noctua gpad file, process it
        if noctua_gpad_file:
            click.echo("Making noctua gpad products...{}".format(noctua_gpad_file))
            # Process noctua gpad file
            process_noctua_gpad_file(noctua_gpad_file, gpadwriter, ontology_graph, gpi)

        # Process the GAF file
        process_gaf_file(gaf_path, gpadwriter, ontology_graph, paint_gaf_src)

    # The file will be automatically closed here, after exiting the 'with' block
    return [gpad_file_path]


def process_noctua_gpad_file(noctua_gpad_file, gpadwriter, ontology_graph, gpi):
    """
    Process a noctua gpad file and write the associations to the gpad writer.

    :param noctua_gpad_file: The path to the noctua gpad file
    :param gpadwriter: The gpad writer to write the associations to
    :param ontology_graph: The ontology graph to use for parsing the associations
    :param gpi: The path to the gpi file -- needed to convert isoform annotations from Noctua files
    """

    with open(noctua_gpad_file) as nf:
        lines = sum(1 for line in nf)
        nf.seek(0)  # Reset file pointer to the beginning after counting lines
        gpadparser = GpadParser(config=assocparser.AssocParserConfig(ontology=ontology_graph,
                                                                     paint=False,
                                                                     rule_set="all"))
        click.echo("Making noctua gpad products...")
        with click.progressbar(iterable=gpadparser.association_generator(file=nf), length=lines) as associations:
            for association in associations:
                # If the association is an isoform annotation, convert it to a gene annotation
                gpadwriter.write_assoc(association)


def process_gaf_file(gaf_path, gpadwriter, ontology_graph, paint_gaf_src):
    """
    Process a gaf file and write the associations to the gpad writer.

    :param gaf_path: The path to the gaf file
    :param gpadwriter: The gpad writer to write the associations to
    :param ontology_graph: The ontology graph to use for parsing the associations
    :param paint_gaf_src: The source of the paint gaf file

    """
    with open(gaf_path) as gf:
        lines = sum(1 for line in gf)
        gf.seek(0)  # Reset file pointer to the beginning after counting lines
        gafparser = GafParser(config=assocparser.AssocParserConfig(ontology=ontology_graph,
                                                                   paint=True,
                                                                   rule_set="all"))
        click.echo("Merging in source gaf to gpad product...")
        with click.progressbar(iterable=gafparser.association_generator(file=gf), length=lines) as associations:
            for association in associations:
                gpadwriter.write_assoc(association)

    if paint_gaf_src is not None:
        with open(paint_gaf_src) as pgf:
            lines = sum(1 for line in pgf)
            pgf.seek(0)
            gafparser = GafParser(config=assocparser.AssocParserConfig(ontology=ontology_graph,
                                                                       paint=True,
                                                                       rule_set="all"))
            click.echo("Merging in paint gaf to gpad product...")
            with click.progressbar(iterable=gafparser.association_generator(file=pgf), length=lines) as associations:
                for association in associations:
                    gpadwriter.write_assoc(association)


@tools.gzips
def produce_gpi(dataset, target_dir, gaf_path, ontology_graph, gpad_gpi_output_version):
    gafparser = GafParser()
    gafparser.config = assocparser.AssocParserConfig(
        ontology=ontology_graph
    )
    with open(gaf_path) as sg:
        lines = sum(1 for line in sg)

    gpi_path = os.path.join(os.path.split(gaf_path)[0], "{}.gpi".format(dataset))
    with open(gaf_path) as gf, open(gpi_path, "w") as gpi:
        click.echo("Using {} as the gaf to build gpi with".format(gaf_path))
        bridge = gafgpibridge
        gpiwriter = entitywriter.GpiWriter(file=gpi, version=gpad_gpi_output_version)
        gpi_cache = set()

        with click.progressbar(iterable=gafparser.association_generator(file=gf), length=lines) as associations:
            for association in associations:
                entity = bridge.convert_association(association)
                if entity not in gpi_cache and entity is not None:
                    # If the entity is not in the cache, add it and write it out
                    gpi_cache.add(entity)
                    gpiwriter.write_entity(entity)
    return gpi_path


@tools.gzips
def produce_ttl(dataset, target_dir, gaf_path, ontology_graph):
    gafparser = GafParser()
    gafparser.config = assocparser.AssocParserConfig(
        ontology=ontology_graph
    )

    with open(gaf_path) as sg:
        lines = sum(1 for line in sg)

    ttl_path = os.path.join(os.path.split(gaf_path)[0], "{}_cam.ttl".format(dataset))
    click.echo("Producing ttl: {}".format(ttl_path))
    rdf_writer = assoc_rdfgen.TurtleRdfWriter()
    transformer = assoc_rdfgen.CamRdfTransform(writer=rdf_writer)
    parser_config = assocparser.AssocParserConfig(ontology=ontology_graph)

    with open(gaf_path) as gf:
        with click.progressbar(iterable=gafparser.association_generator(file=gf), length=lines) as associations:
            for association in associations:
                if "header" not in association or not association["header"]:
                    transformer.provenance()
                    transformer.translate(association)

    with open(ttl_path, "wb") as ttl:
        click.echo("Writing ttl to disk")
        rdf_writer.serialize(destination=ttl)

    return ttl_path


@tools.gzips
def merge_all_mixin_gaf_into_mod_gaf(valid_gaf_path, mixin_gaf_paths):
    def header_and_annotations(gaf_file):
        headers = []
        annotations = []

        for line in gaf_file.readlines():
            line = line.rstrip("\n")
            if line.startswith("!"):
                headers.append(line)
            else:
                annotations.append(line)

        return (headers, annotations)

    def make_mixin_header(header_lines, path):
        the_header = [
                         "!Header copied from {}".format(os.path.basename(path)),
                         "!================================="
                     ] + header_lines[8:] + ["!"]
        return the_header

    # ####################################################################

    # Set up merged final gaf product path
    dirs, name = os.path.split(valid_gaf_path)
    merged_path = os.path.join(dirs, "{}.gaf".format(name.rsplit("_", maxsplit=1)[0]))
    valid_header = []
    annotations = []
    with open(valid_gaf_path) as valid_file:
        valid_header, annotations = header_and_annotations(valid_file)

    mixin_headers = []
    for mixin_gaf_path in mixin_gaf_paths:
        with open(mixin_gaf_path) as mixin_file:
            mixin_header, mixin_annotations = header_and_annotations(mixin_file)

            mixin_headers += make_mixin_header(mixin_header, mixin_gaf_path)
            annotations += mixin_annotations

    full_header = valid_header + \
                  [
                      "!=================================",
                      "!"
                  ] + mixin_headers + \
                  [
                      "!=================================",
                      "!",
                      "!Documentation about this header can be found here: https://github.com/geneontology/go-site/blob/master/docs/gaf_validation.md",
                      "!"
                  ]
    all_lines = full_header + annotations

    with open(merged_path, "w") as merged_file:
        merged_file.write("\n".join(all_lines))

    return merged_path


def mixin_a_dataset(valid_gaf, mixin_metadata_list, group_id, dataset, target, ontology, gpipaths=None,
                    base_download_url=None, rule_metadata={}, replace_existing_files=True, rule_contexts=[],
                    gaf_output_version="2.2"):
    end_gaf = valid_gaf
    mixin_gaf_paths = []
    for mixin_metadata in mixin_metadata_list:
        click.echo("Merging mixin dataset {}".format(mixin_metadata["id"]))
        mixin_src = check_and_download_mixin_source(mixin_metadata, group_id, dataset, target,
                                                    base_download_url=base_download_url,
                                                    replace_existing_files=replace_existing_files)

        if mixin_src is not None:
            mixin_dataset_metadata = mixin_dataset(mixin_metadata, dataset)
            mixin_dataset_id = mixin_dataset_metadata["dataset"]
            format = mixin_dataset_metadata["type"]
            context = ["import"] if mixin_metadata.get("import", False) else []
            mixin_gaf = produce_gaf(mixin_dataset_id, mixin_src, ontology, gpipaths=gpipaths, paint=True,
                                    group=mixin_metadata["id"], rule_metadata=rule_metadata, format=format,
                                    rule_contexts=context, gaf_output_version=gaf_output_version)[0]
            mixin_gaf_paths.append(mixin_gaf)

    if mixin_gaf_paths:
        # If we found and processed any mixin gafs, then lets merge them.
        end_gaf = merge_all_mixin_gaf_into_mod_gaf(valid_gaf, mixin_gaf_paths)
    else:
        gafgz = "{}.gz".format(valid_gaf)
        shutil.copyfile(gafgz, os.path.join(os.path.split(gafgz)[0], "{}.gaf.gz".format(dataset)))

    return end_gaf


@click.group()
@click.option("--verbose", "-v", is_flag=True, default=False)
@click.pass_context
def cli(ctx, verbose):
    if verbose:
        logger.setLevel(logging.INFO)


@cli.command()
@click.pass_context
@click.argument("group")
@click.option("--metadata", "-m", "metadata_dir", type=click.Path(), required=True)
@click.option("--gpad", default=False, is_flag=True)
@click.option("--gpad-gpi-output-version", default="1.2", type=click.Choice(["1.2", "2.0"]))
@click.option("--ttl", default=False, is_flag=True)
@click.option("--target", "-t", type=click.Path(), required=True)
@click.option("--ontology", "-o", type=click.Path(exists=True), required=False)
@click.option("--exclude", "-x", multiple=True)
@click.option("--base-download-url", "-b", default=None)
@click.option("--suppress-rule-reporting-tag", "-S", multiple=True,
              help="Suppress markdown output messages from rules tagged with this tag")
@click.option("--skip-existing-files", "-K", is_flag=True, default=False,
              help="When downloading files, if a file already exists it won't downloaded over")
@click.option("--gaferencer-file", "-I", type=click.Path(exists=True), default=None, required=False,
              help="Path to Gaferencer output to be used for inferences")
@click.option("--only-dataset", default=None)
@click.option("--gaf-output-version", default="2.2", type=click.Choice(["2.1", "2.2"]))
@click.option("--rule-set", "-l", "rule_set", default=[assocparser.RuleSet.ALL], multiple=True)
@click.option("--retracted_pub_set", type=click.Path(exists=True), default=None, required=False,
              help="Path to retracted publications file")
def produce(ctx, group, metadata_dir, gpad, gpad_gpi_output_version, ttl, target, ontology, exclude, base_download_url,
            suppress_rule_reporting_tag, skip_existing_files, gaferencer_file, only_dataset, gaf_output_version,
            rule_set, retracted_pub_set):
    """
    Produce GAF, GPI, and TTL files for a group.

    This command will download the GAF files for a group, validate them, and then produce GPI and TTL files.
    :param ctx: Click context
    :param group: The group to produce files for
    :param metadata_dir: The directory containing the metadata files
    :param gpad: Produce GPAD files
    :param gpad_gpi_output_version: The version of the GPAD and GPI files to produce
    :param ttl: Produce TTL files
    :param target: The directory to put the files in
    :param ontology: The ontology to use for validation
    :param exclude: Datasets to exclude
    :param base_download_url: The base URL to download files from
    :param suppress_rule_reporting_tag: Tags to suppress in the rule reporting
    :param skip_existing_files: Skip downloading files that already exist
    :param gaferencer_file: The path to the Gaferencer output file
    :param only_dataset: Only process a single dataset
    :param gaf_output_version: The version of the GAF files to produce
    :param rule_set: The rule set to use
    :param retracted_pub_set: The path to the retracted publications file
    """
    logger.info("Logging is verbose")
    products = {
        "gaf": True,
        "gpi": True,
        "gpad": gpad,
        "ttl": ttl
    }
    click.echo("Making products {}.".format(", ".join([key for key in products if products[key]])))
    absolute_target = os.path.abspath(target)
    os.makedirs(os.path.join(absolute_target, "groups"), exist_ok=True)
    click.echo("Products will go in {}".format(absolute_target))
    absolute_metadata = os.path.abspath(metadata_dir)

    group_metadata = metadata.dataset_metadata_file(absolute_metadata, group)
    click.echo("Loading ontology: {}...".format(ontology))
    ontology_graph = OntologyFactory().create(ontology, ignore_cache=True)

    downloaded_gaf_sources = download_source_gafs(group_metadata, absolute_target, exclusions=exclude,
                                                  base_download_url=base_download_url,
                                                  replace_existing_files=not skip_existing_files,
                                                  only_dataset=only_dataset)

    # extract the titles for the go rules, this is a dictionary comprehension
    rule_metadata = metadata.yamldown_lookup(os.path.join(absolute_metadata, "rules"))
    goref_metadata = metadata.yamldown_lookup(os.path.join(absolute_metadata, "gorefs"))
    ref_species_metadata = metadata.yaml_set(absolute_metadata, "go-reference-species.yaml", "taxon_id")

    click.echo("Found {} GO Rules".format(len(rule_metadata.keys())))
    click.echo("Found {} GO_REFs".format(len(goref_metadata.keys())))
    click.echo("Found {} Reference Species".format(len(ref_species_metadata)))

    paint_metadata = metadata.dataset_metadata_file(absolute_metadata, "paint")
    noctua_metadata = metadata.dataset_metadata_file(absolute_metadata, "noctua")
    mixin_metadata_list = list(filter(lambda m: m != None, [paint_metadata]))

    db_entities = metadata.database_entities(absolute_metadata)
    group_ids = metadata.groups(absolute_metadata)
    extensions_constraints = metadata.extensions_constraints_file(absolute_metadata)

    gaferences = None
    if gaferencer_file:
        gaferences = gaference.load_gaferencer_inferences_from_file(gaferencer_file)

    # Default comes through as single-element tuple
    if rule_set == (assocparser.RuleSet.ALL,):
        rule_set = assocparser.RuleSet.ALL

    db_type_name_regex_id_syntax = metadata.database_type_name_regex_id_syntax(absolute_metadata)

    if retracted_pub_set:
        retracted_pubs = metadata.retracted_pub_set(retracted_pub_set)
    else:
        retracted_pubs = metadata.retracted_pub_set_from_meta(absolute_metadata)

    for dataset_metadata, source_gaf in downloaded_gaf_sources:
        dataset = dataset_metadata["dataset"]
        # Set paint to True when the group is "paint".
        # This will prevent filtering of IBA (GO_RULE:26) when paint is being treated as a top level group,
        # like for paint_other.
        click.echo("Producing GAF by passing through validation rules... {}".format(dataset))
        valid_gaf = produce_gaf(dataset, source_gaf, ontology_graph,
                                paint=(group == "paint"),
                                group=group,
                                rule_metadata=rule_metadata,
                                goref_metadata=goref_metadata,
                                ref_species_metadata=ref_species_metadata,
                                db_type_name_regex_id_syntax=db_type_name_regex_id_syntax,
                                retracted_pub_set=retracted_pubs,
                                db_entities=db_entities,
                                group_idspace=group_ids,
                                suppress_rule_reporting_tags=suppress_rule_reporting_tag,
                                annotation_inferences=gaferences,
                                group_metadata=group_metadata,
                                extensions_constraints=extensions_constraints,
                                rule_contexts=["import"] if dataset_metadata.get("import", False) else [],
                                gaf_output_version=gaf_output_version,
                                rule_set=rule_set
                                )[0]

        gpi_list = []

        matching_gpi_path = None
        click.echo("Try to find other GPIs in metadata and merge...")

        for ds in group_metadata["datasets"]:
            # Where type=GPI for the same dataset (e.g. "zfin", "goa_cow")
            if ds["type"] == "gpi" and ds["dataset"] == dataset and ds.get("source"):
                matching_gpi_path = download_a_dataset_source(group, ds, absolute_target, ds["source"],
                                                              replace_existing_files=not skip_existing_files)
                if ds.get("compression", None) == "gzip":
                    matching_gpi_path = unzip_simple(matching_gpi_path)
                gpi_list.append(matching_gpi_path)

        click.echo("Found the matching gpi path...{}".format(matching_gpi_path))

        click.echo("Downloading the noctua and paint GPAD files...")
        noctua_gpad_src = check_and_download_mixin_source(noctua_metadata, group_metadata["id"], dataset, target,
                                                          base_download_url=base_download_url,
                                                          replace_existing_files=not skip_existing_files)
        paint_gaf_src = (check_and_download_mixin_source(paint_metadata, group_metadata["id"], dataset, target,
                                                         base_download_url=base_download_url,
                                                         replace_existing_files=not skip_existing_files)
                         if paint_metadata else None)

        click.echo("Producing GPI for use in creating GPADs...")
        gpi = produce_gpi(dataset, absolute_target, valid_gaf, ontology_graph, gpad_gpi_output_version)
        gpi_list.append(gpi)
        click.echo("Executing 'make_gpads' in validate.produce with all the assembled GAF files...")
        make_gpads(dataset, valid_gaf, products,
                   ontology_graph, noctua_gpad_src, paint_gaf_src,
                   gpi, gpad_gpi_output_version)

        end_gaf = mixin_a_dataset(valid_gaf, [noctua_metadata, paint_metadata],
                                  group_metadata["id"], dataset, absolute_target,
                                  ontology_graph, gpipaths=gpi_list, base_download_url=base_download_url,
                                  rule_metadata=rule_metadata, replace_existing_files=not skip_existing_files,
                                  gaf_output_version=gaf_output_version)

        click.echo("Pre-isoform fix gaf file...{}".format(end_gaf))
        click.echo("Executing the isoform fixing step in validate.produce...")
        # run the resulting gaf through one last parse and replace, to handle the isoforms
        # see: https://github.com/geneontology/go-site/issues/2291
        temp_output_gaf_path = os.path.join(os.path.split(end_gaf)[0], "{}_temp.gaf".format(dataset))
        click.echo("temp_output_gaf_path: {}".format(temp_output_gaf_path))
        isoform_fixed_gaf = fix_pro_isoforms_in_gaf(end_gaf, matching_gpi_path, ontology_graph, temp_output_gaf_path)
        click.echo("isoform_fixed_gaf: ".format(isoform_fixed_gaf))

        final_output_gaf_path = os.path.join(os.path.split(end_gaf)[0], "{}.gaf".format(dataset))

        click.echo("Rename the temporary isoform fixed file to the final GAF...")
        os.rename(temp_output_gaf_path, final_output_gaf_path)
        click.echo("final_output_gaf_path: ".format(final_output_gaf_path))

        click.echo("Producing final GPI after all GAF corrections...")
        final_gpi = produce_gpi(dataset, absolute_target, final_output_gaf_path, ontology_graph, gpad_gpi_output_version)

        click.echo("Creating ttl files...")
        make_ttls(dataset, final_output_gaf_path, products, ontology_graph)


def fix_pro_isoforms_in_gaf(gaf_file_to_fix: str,
                            gpi_file: str,
                            ontology_graph,
                            output_file_path: str) -> str:
    """
    Given a GAF file and a GPI file, fix the GAF file by converting isoform annotations to gene annotations. Storing
    the isoforms back in subject_extensions collection, changing the full_name, synonyms, label, and type back to the
    gene in the GPI file.
    :param gaf_file_to_fix: The path to the GAF file to fix
    :param gpi_file: The path to the GPI file
    :param ontology_graph: The ontology graph to use for parsing the associations
    :param output_file_path: The path to write the fixed GAF file to
    :return: The path to the fixed GAF file
    """
    fixed_associations = []
    gpiparser = GpiParser(config=assocparser.AssocParserConfig(ontology=ontology_graph))
    # Parse the GPI file, creating a map of identifiers to GPI entries
    gpis = gpiparser.parse(gpi_file, None)
    gpi_map = {}
    for gpi_entry in gpis:
        gpi_map[gpi_entry.get('id')] = {"encoded_by": gpi_entry.get('encoded_by'),
                                        "full_name": gpi_entry.get('full_name'),
                                        "label": gpi_entry.get('label'),
                                        "synonyms": gpi_entry.get('synonyms'),
                                        # GPI spec says this is single valued, but GpiParser returns this as a list.
                                        "type": gpi_entry.get('type')[0],
                                        "id": gpi_entry.get('id')}

    gafparser = GafParser(config=assocparser.AssocParserConfig(ontology=ontology_graph))
    gafwriter = GafWriter(file=open(output_file_path, "w"), version="2.2")

    # these are statistic parameters that record when a substitution is made.
    substitution_count = 0
    no_substitution_count = 0

    with open(gaf_file_to_fix, "r") as file:
        for line in file:
            annotation = gafparser.parse_line(line)
            for source_assoc in annotation.associations:
                if isinstance(source_assoc, dict):
                    continue  # skip the header
                if source_assoc.subject.id.namespace.startswith("PR"):
                    full_old_identifier = source_assoc.subject.id.namespace + ":" + source_assoc.subject.id.identity
                    old_namespace = source_assoc.subject.id.namespace
                    old_identity = source_assoc.subject.id.identity
                    # TODO: right now we get the FIRST encoded_by result -- this is what the original script from Chris did??
                    if "MGI" == gpi_map[full_old_identifier].get("encoded_by")[0].split(":")[0]:
                        source_assoc.subject.id.namespace = gpi_map[full_old_identifier].get("encoded_by")[0].split(":")[0]
                        source_assoc.subject.id.identity = "MGI:" + gpi_map[full_old_identifier].get("encoded_by")[0].split(":")[2]
                    else:
                        source_assoc.subject.id.namespace = \
                        gpi_map[full_old_identifier].get("encoded_by")[0].split(":")[0]
                        source_assoc.subject.id.identity = \
                        gpi_map[full_old_identifier].get("encoded_by")[0].split(":")[1]
                    full_new_identifier = source_assoc.subject.id.namespace + ":" + source_assoc.subject.id.identity
                    source_assoc.subject.full_name = gpi_map[full_new_identifier].get("full_name")
                    source_assoc.subject.label = gpi_map[full_new_identifier].get("label")
                    source_assoc.subject.synonyms = gpi_map[full_new_identifier].get("synonyms")
                    source_assoc.subject.type = gpi_map[full_new_identifier].get("type")

                    # we need to put the isoform currently being swapped, back into "Column 17" which is a
                    # subject_extension member.
                    isoform_term = Curie(namespace=old_namespace, identity=old_identity)
                    isoform_relation = Curie(namespace="RO", identity="0002327")
                    new_subject_extension = ExtensionUnit(relation=isoform_relation, term=isoform_term)
                    source_assoc.subject_extensions.append(new_subject_extension)

                    # count the substitution here for reporting later
                    substitution_count += 1
                else:
                    no_substitution_count += 1

                # Join fields back into a string and write to output file
                fixed_associations.append(source_assoc)

    gafwriter.write(fixed_associations)
    click.echo(f"Substituted {substitution_count} entries in {gaf_file_to_fix} "
               f"and left {no_substitution_count} entries unchanged.")

    return output_file_path

@cli.command()
@click.pass_context
@click.option("--gpad_path", "-g", type=click.Path(), required=True)
@click.option("--gpi_path", "-i", type=click.Path(), required=True)
@click.option("--target", "-t", type=click.Path(), required=True)
@click.option("--ontology", "-o", type=click.Path(exists=True), required=True, multiple=True)
@click.option("--ttl", default=False, is_flag=True)
@click.option("--modelstate", "-s", default=None)
def gpad2gocams(ctx, gpad_path, gpi_path, target, ontology, ttl, modelstate):
    # NOTE: Validation on GPAD not included here since it's currently baked into produce() above.
    # Multi-param to accept multiple ontology files, then merge to one (this will make a much smaller ontology
    #  with only what we need, i.e. GO, RO, GOREL)
    ontology_graph = OntologyFactory().create(ontology[0], ignore_cache=True)
    for ont in ontology[1:]:
        ontology_graph.merge([OntologyFactory().create(ont, ignore_cache=True)])
    parser_config = assocparser.AssocParserConfig(ontology=ontology_graph,
                                                  gpi_authority_path=gpi_path
                                                  )
    extractor = AssocExtractor(gpad_path, parser_config=parser_config)
    assocs_by_gene = extractor.group_assocs()

    absolute_target = os.path.abspath(target)
    gpad_basename = os.path.basename(gpad_path)
    gpad_basename_root, gpad_ext = os.path.splitext(gpad_basename)
    output_basename = "{}.nq".format(gpad_basename_root)
    report_basename = "{}.gocamgen.report".format(gpad_basename_root)
    output_path = os.path.join(absolute_target, output_basename)
    report_path = os.path.join(absolute_target, report_basename)

    builder = GoCamBuilder(parser_config=parser_config, modelstate=modelstate)

    for gene, associations in assocs_by_gene.items():
        if ttl:
            builder.make_model_and_write_out(gene, annotations=associations, output_directory=absolute_target)
        else:
            builder.make_model_and_add_to_store(gene, annotations=associations)
    if not ttl:
        builder.write_out_store_to_nquads(filepath=output_path)

    builder.write_report(report_filepath=report_path)


@cli.command()
@click.argument("group")
@click.argument("dataset")
@click.option("--metadata", "-m", type=click.Path(), required=True)
@click.option("--target", type=click.Path(), required=True)
@click.option("--ontology", type=click.Path(), required=True)
def paint(group, dataset, metadata, target, ontology):
    absolute_metadata = os.path.abspath(metadata)
    absolute_target = os.path.abspath(target)
    os.makedirs(os.path.join(absolute_target, "groups"), exist_ok=True)
    paint_metadata = metadata.dataset_metadata_file(absolute_metadata, "paint")

    paint_src_gaf = check_and_download_mixin_source(paint_metadata, dataset, absolute_target)

    click.echo("Loading ontology: {}...".format(ontology))
    ontology_graph = OntologyFactory().create(ontology)

    gpi_path = os.path.join(absolute_target, "groups", dataset, "{}.gpi".format(dataset))
    click.echo("Using GPI at {}".format(gpi_path))
    paint_gaf = produce_gaf("paint_{}".format(dataset), paint_src_gaf, ontology_graph, gpipath=gpi_path)


@cli.command()
@click.option("--metadata", "-m", "metadata_dir", type=click.Path(), required=True)
@click.option("--out", type=click.Path(), required=False)
@click.option("--ontology", type=click.Path(), required=True)
@click.option("--gaferencer-file", "-I", type=click.Path(exists=True), default=None, required=False,
              help="Path to Gaferencer output to be used for inferences")
@click.option("--retracted_pub_set", type=click.Path(exists=True), default=None, required=False,
              help="Path to retracted publications file")
def rule(metadata_dir, out, ontology, gaferencer_file, retracted_pub_set):
    absolute_metadata = os.path.abspath(metadata_dir)

    click.echo("Loading ontology: {}...".format(ontology))
    ontology_graph = OntologyFactory().create(ontology)

    goref_metadata = metadata.yamldown_lookup(os.path.join(absolute_metadata, "gorefs"))
    gorule_metadata = metadata.yamldown_lookup(os.path.join(absolute_metadata, "rules"))
    ref_species_metadata = metadata.yaml_set(absolute_metadata, "go-reference-species.yaml", "taxon_id")
    db_type_name_regex_id_syntax = metadata.database_type_name_regex_id_syntax(absolute_metadata)
    retracted_pubs = None
    if retracted_pub_set:
        retracted_pubs = metadata.retracted_pub_set(retracted_pub_set)
    else:
        retracted_pubs = metadata.retracted_pub_set_from_meta(absolute_metadata)

    click.echo("Found {} GO Rules".format(len(gorule_metadata.keys())))

    db_entities = metadata.database_entities(absolute_metadata)
    group_ids = metadata.groups(absolute_metadata)

    gaferences = None
    if gaferencer_file:
        gaferences = gaference.load_gaferencer_inferences_from_file(gaferencer_file)

    config = assocparser.AssocParserConfig(
        ontology=ontology_graph,
        goref_metadata=goref_metadata,
        ref_species_metadata=ref_species_metadata,
        db_type_name_regex_id_syntax=db_type_name_regex_id_syntax,
        retracted_pub_set=retracted_pubs,
        entity_idspaces=db_entities,
        group_idspace=group_ids,
        annotation_inferences=gaferences,
        rule_set=assocparser.RuleSet.ALL
    )
    all_examples_valid = True
    all_results = []
    for rule_id, rule_meta in gorule_metadata.items():
        examples = rules.RuleExample.example_from_json(rule_meta)
        if len(examples) == 0:
            # skip if there are no examples
            continue

        click.echo("==============================================================================")
        click.echo("Validating {} examples for {}".format(len(examples), rule_id.upper().replace("-", ":")))
        results = rules.validate_all_examples(examples, config=config)
        successes = sum(1 for r in results if r.success)
        click.echo("\t* {}/{} success".format(successes, len(results)))
        for r in results:
            if not r.success:
                click.echo("\tRule example failed: {}".format(r.reason))
                click.echo("\tInput: >> `{}`".format(r.example.input))
                all_examples_valid = False

        all_results += results

    if out:
        absolute_out = os.path.abspath(out)
        os.makedirs(os.path.dirname(absolute_out), exist_ok=True)
        try:
            with open(absolute_out, "w") as outfile:
                json.dump(rules.validation_report(all_results), outfile, indent=4)
        except Exception as e:
            raise click.ClickException("Could not write report to {}: ".format(out, e))

    if not all_examples_valid:
        raise click.ClickException("At least one rule example was not validated.")


if __name__ == "__main__":
    cli(obj={})



from pathlib import Path

from bin import validate
from ontobio.ontol_factory import OntologyFactory
from pprint import pprint

gpi_file = "mgi.truncated.gpi2"
gaf_file_to_fix = "mgi.gaf"
output_file_path = "fixed_test.gaf"
ontology = "go"
ontology_graph = OntologyFactory().create(ontology, ignore_cache=True)


def test_fix_isoforms():
    base_path = Path(__file__).parent / "resources"
    gpi_path = base_path / gpi_file
    # Define the paths for the GAF and expected GPI file
    gaf_path = base_path / gaf_file_to_fix

    # Ensure the GAF file exists to avoid FileNotFoundError
    assert gaf_path.exists()
    assert gpi_path.exists()

    validate.fix_pro_isoforms_in_gaf(str(gaf_path), str(gpi_path), ontology_graph, output_file_path)
    with open(output_file_path, 'r') as f:
        lines = f.readlines()
    for line in lines:
        assert not line.startswith("PR")
    assert len(lines) > 100




from pathlib import Path

from ontobio.model import collections
from ontobio.model.association import Curie, Subject
from ontobio.io import assocparser, gafparser, gpadparser


def test_bioentities_get_when_empty():
    e = collections.BioEntities(dict())
    assert e.get(Curie("FOO", "123")) == None


def test_bioentities_get_when_exists():
    e = collections.BioEntities({
        Curie("FOO", "123"): Subject(Curie("FOO", "123"), "hello", "world", [], "protien", Curie("NCBITaxon", "12345"))
    })
    assert e.get(Curie.from_str("FOO:123")) == Subject(Curie("FOO", "123"), "hello", "world", [], "protien",
                                                       Curie("NCBITaxon", "12345"))


def test_bioentities_merge():
    e = collections.BioEntities({
        Curie("FOO", "123"): Subject(Curie("FOO", "123"), "hello", "world", [], "protien", Curie("NCBITaxon", "12345"))
    })

    o = collections.BioEntities({
        Curie("BAR", "987"): Subject(Curie("BAR", "987"), "goodbye", "world", [], "protien", Curie("NCBITaxon", "999"))
    })

    assert e.merge(o) == collections.BioEntities({
        Curie("FOO", "123"): Subject(Curie("FOO", "123"), "hello", "world", [], "protien", Curie("NCBITaxon", "12345")),
        Curie("BAR", "987"): Subject(Curie("BAR", "987"), "goodbye", "world", [], "protien", Curie("NCBITaxon", "999"))
    })

    # Test that blank label does not overwrite existing label
    p = collections.BioEntities({
        Curie("BAR", "987"): Subject(Curie("BAR", "987"), "", "world", [], "protien", Curie("NCBITaxon", "999"))
    })

    assert o.merge(p) == collections.BioEntities({
        Curie("BAR", "987"): Subject(Curie("BAR", "987"), "goodbye", "world", [], "protien", Curie("NCBITaxon", "999"))
    })


def test_bioentities_merge_clobber():
    e = collections.BioEntities({
        Curie("FOO", "123"): Subject(Curie("FOO", "123"), "hello", "world", [], "protien", Curie("NCBITaxon", "12345"))
    })

    o = collections.BioEntities({
        Curie("FOO", "123"): Subject(Curie("FOO", "123"), "different", "world", [], "dog", Curie("NCBITaxon", "12345"))
    })
    # Get the clobbered key, the value should be the subject in `o`
    assert e.merge(o).get(Curie("FOO", "123")) == Subject(Curie("FOO", "123"), "different", "world", [], "dog",
                                                          Curie("NCBITaxon", "12345"))


def test_bioentities_load_from_file():
    pombase = collections.BioEntities.load_from_file("tests/resources/truncated-pombase.gpi")
    assert len(pombase.entities.keys()) == 199  # Has 199 gpi lines in the file
    assert pombase.get(Curie.from_str("PomBase:SPAC1565.04c")) == Subject(Curie.from_str("PomBase:SPAC1565.04c"),
                                                                          "ste4", ["adaptor protein Ste4"], [],
                                                                          ["protein"], Curie.from_str("NCBITaxon:4896"))


def test_bioentities_file_gone():
    gone = collections.BioEntities.load_from_file("not_here.gpi")
    assert gone.entities == dict()


def test_create_parser_non_version_line():
    parser = collections.create_parser_from_header("! hello", assocparser.AssocParserConfig())
    assert parser == None


def test_create_parser_gaf():
    parser = collections.create_parser_from_header("!gaf-version: 2.1", assocparser.AssocParserConfig())
    assert isinstance(parser, gafparser.GafParser)
    assert parser.version == "2.1"

    parser = collections.create_parser_from_header("!gaf-version: 2.2", assocparser.AssocParserConfig())
    assert isinstance(parser, gafparser.GafParser)
    assert parser.version == "2.2"


def test_create_parser_gpad():
    parser = collections.create_parser_from_header("!gpa-version: 1.2", assocparser.AssocParserConfig())
    assert isinstance(parser, gpadparser.GpadParser)
    assert parser.version == "1.2"

    parser = collections.create_parser_from_header("!gpad-version: 2.0", assocparser.AssocParserConfig())
    assert isinstance(parser, gpadparser.GpadParser)
    assert parser.version == "2.0"


def test_construct_collection_empty():
    collection = collections.construct_collection(None, [], assocparser.AssocParserConfig())
    assert collection.headers == []
    assert collection.associations == collections.GoAssociations([])
    assert collection.entities == collections.BioEntities(dict())


def test_construct_collection_basic():
    collection = collections.construct_collection("tests/resources/truncated-pombase.gaf",
                                                  ["tests/resources/truncated-pombase.gpi"],
                                                  assocparser.AssocParserConfig())
    # There are 24 lines of headers in the truncated-pombase.gaf
    assert len(collection.headers) == 25
    # There are many lines in this file, we should get more than 100 back
    assert len(collection.associations.associations) > 100
    # This is a 199 line gpi file, with each identifier unique, so should have 200 entity entries
    assert len(collection.entities.entities.keys()) == 199


def test_construct_collection_no_version_error():
    collection = collections.construct_collection("tests/resources/no-version.gaf", [], assocparser.AssocParserConfig())
    assert collection.associations.associations == []
    assert collection.report.to_report_json()["messages"]["gorule-0000001"][0]["type"] == "Invalid Annotation File"


def test_bioentities_from_gpi_2_0():
    base_path = Path(__file__).parent / "resources"
    gpi_path = base_path / "mgi.truncated.gpi2"
    entities = collections.BioEntities.load_from_file(str(gpi_path))
    assert entities.get(Curie(namespace="MGI", identity="MGI:1918925")) == Subject(
        id=Curie.from_str("MGI:MGI:1918925"),
        label="Sanbr",
        fullname=["SANT and BTB domain regulator of CSR"],
        synonyms=["0610010F05Rik"],
        type=[Curie(namespace='SO', identity='0001217')],
        taxon=Curie(namespace="NCBITaxon", identity="10090"),
        db_xrefs=[Curie(namespace='UniProtKB', identity='Q68FF0')])



"""
Dataclasses for ontobio.sim
"""
from dataclasses import dataclass, field
from typing import List, Optional, Union


@dataclass
class Node:
    """
    Basic node
    """
    id: str
    label: str = None


@dataclass
class ICNode(Node):
    """
    Node with information content
    """
    IC: float = None


@dataclass
class TypedNode(Node):
    """
    Node with type and optional taxon
    """
    type: str = None
    taxon: Optional[Node] = None


@dataclass
class SimMetadata():
    """
    Metadata returned with sim result
    """
    max_max_ic: float


@dataclass
class PairwiseMatch:
    """
    Data class for pairwise match
    """
    reference: ICNode
    match: ICNode
    lcs: ICNode


@dataclass
class SimMatch(TypedNode):
    """
    Data class similarity match
    """
    rank: Union[int, str] = None
    score: Union[float, int] = None
    significance: Union[float, str, None] = None
    pairwise_match: List[PairwiseMatch] = field(default_factory=list)


@dataclass
class SimQuery:
    """
    Data class for query
    """
    ids: List[Node]
    negated_ids: List[Node] = field(default_factory=list)
    unresolved_ids: List[str] = field(default_factory=list)
    target_ids: List[List[Node]] = field(default_factory=lambda: [[]])
    reference: TypedNode = None


@dataclass
class SimResult:
    """
    Data class similarity result
    """
    query: SimQuery
    matches: List[SimMatch]
    metadata: SimMetadata


@dataclass
class IcStatistic:
    """
    Data class for ic statistics

    mean_mean_ic: average of the average IC per individual
    mean_sum_ic: average sumIC per individual
    mean_cls: float, avg number of classes per individual
    max_max_ic: maximum IC of class annotated to an individual
    max_sum_ic: maximum sumIC
    individual_count: number of individuals
    mean_max_ic: average maxIC per individual
    """
    mean_mean_ic: float
    mean_sum_ic: float
    mean_cls: float
    max_max_ic: float
    max_sum_ic: float
    individual_count: int
    mean_max_ic: float
    descendants: List[str] = None


@dataclass
class AnnotationSufficiency:
    """
    Data class for annotation sufficiency object
    """
    simple_score: float
    scaled_score: float
    categorical_score: float



import os
import sys
import re
import subprocess

import setuptools

if sys.version_info.major >= 3.10:
    sys.exit("Error: Python 3.10 or later is required")

directory = os.path.dirname(os.path.abspath(__file__))

# version
init_path = os.path.join(directory, 'ontobio', '__init__.py')
with open(init_path) as read_file:
    text = read_file.read()
pattern = re.compile(r"^__version__ = ['\"]([^'\"]*)['\"]", re.MULTILINE)
version = pattern.search(text).group(1)

with open("requirements.txt", "r") as FH:
    REQUIREMENTS = FH.readlines()

setuptools.setup(
    name='ontobio',
    version=version,
    author='Chris Mungall',
    author_email='cmungall@gmail.com',
    url='https://github.com/biolink/ontobio',
    description='Library for working with OBO Library Ontologies and associations',
    long_description=open("README.rst").read(),
    license='BSD',
    packages=setuptools.find_packages(),
    package_data={"ontobio": ["ontobio/config.yaml"]},

    keywords='ontology graph obo owl sparql networkx network',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Visualization'
    ],

    include_package_data=True,

    # Dependencies
    install_requires=[r for r in REQUIREMENTS if not r.startswith("#")],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['plotly'],
        'test': ['pytest'],
    },
    scripts=[
        'bin/ogr.py',
        'bin/ontobio-assoc.py',
        'bin/ontobio-parse-assocs.py',
        'bin/ontobio-lexmap.py',
        'bin/rdfgen.py',
        'bin/validate.py',
        'bin/materialize.py',
        'bin/clear-cache.py'
    ]
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    #entry_points={
    #    'console_scripts': [
    #        'sample=sample:main',
    #    ],
    #},
)



from ontobio.io import assocparser
from ontobio.io.gpadparser import GpadParser
from ontobio.io import gafparser, gafgpibridge
from ontobio.io.gafparser import GafParser
from ontobio.io import GafWriter
from ontobio.io.assocwriter import GpadWriter
from ontobio.assoc_factory import AssociationSetFactory
from ontobio.ontol_factory import OntologyFactory
from ontobio.model import association
from ontobio.rdfgen import relations
from ontobio.model.association import ConjunctiveSet, ExtensionUnit, Curie
from ontobio.ecomap import EcoMap

import tempfile
import logging
import pytest
import io
import json
import re

ecomap = EcoMap()
ecomap.mappings()

POMBASE = "tests/resources/truncated-pombase.gaf"
POMBASE_GPAD = "tests/resources/truncated-pombase.gpad"
ONT = "tests/resources/go-truncated-pombase.json"
QGAF = "tests/resources/test-qualifiers.gaf"
ALT_ID_ONT = "tests/resources/obsolete.json"
ZFIN_GAF = "tests/resources/zfin_test_gaf.gaf"
OBSOLETE_ONT = "tests/resources/obsolete.json"


def test_skim_gaf():
    p = GafParser()
    p.config.ecomap = EcoMap()
    results = p.skim(open(POMBASE, "r"))
    assert len(results) == 370
    for r in results:
        print(str(r))
        (s, sn, o) = r
        assert o.startswith('GO:')
        assert s.startswith('PomBase:')


def test_skim_gaf_qualifiers():
    p = GafParser()
    p.config.ecomap = EcoMap()
    p.config.remove_double_prefixes = True
    results = p.skim(open(QGAF, "r"))
    for r in results:
        print(str(r))
        (s, sn, o) = r
        assert o.startswith('GO:')
        assert s.startswith('MGI:') or s.startswith('PomBase')
    assert len(results) == 5  # ensure NOTs are skipped

    p.config.exclude_relations = ['contributes_to', 'colocalizes_with']
    results = p.skim(open(QGAF, "r"))
    for r in results:
        (s, sn, o) = r
        assert o.startswith('GO:')
        assert s.startswith('MGI:') or s.startswith('PomBase')
    assert len(results) == 3 # ensure NOTs and excludes relations skipped


def test_one_line():
    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create("tests/resources/goslim_generic.json")))

    parsed = p.parse_line("PomBase	SPBC16D10.09	pcn1		GO:0009536	PMID:8663159	IDA		C	PCNA	pcn	protein	taxon:4896	20150326	PomBase")


def test_skim_gpad():
    p = GpadParser()
    p.config.ecomap = EcoMap()
    results = p.skim(open(POMBASE_GPAD, "r"))
    assert len(results) == 1984
    for r in results:
        print(str(r))
        (s, sn, o) = r
        assert o.startswith('GO:')
        assert s.startswith('PomBase:') or s.startswith('PR:')


def test_parse_gaf():
    parse_with(POMBASE, GafParser())


def test_parse_gpad():
    parse_with(POMBASE_GPAD, GpadParser())


def parse_with(f, p):
    p.config.ecomap = EcoMap()
    is_gaf = f == POMBASE
    ont = OntologyFactory().create(ONT)

    if is_gaf:
        # only do ontology checking on GAF parse;
        # this is because ontology is made from GAF
        p.config.ontology = ont
    else:
        p.config.ontology = None

    results = p.parse(open(f, "r"), skipheader=True)
    print(p.report.to_markdown())
    r1 = results[0]
    # TODO: test datafile does not have ECOs yet!!
    assert ecomap.ecoclass_to_coderef(str(r1.evidence.type))[0] == 'ISO' or str(r1.evidence.type) == 'ECO:0000201'
    assert r1.evidence.with_support_from == [association.ConjunctiveSet([association.Curie.from_str('SGD:S000001583')])]
    assert r1.evidence.has_supporting_reference == [association.Curie.from_str('GO_REF:0000024')]

    if is_gaf:
        assert r1.subject.label == 'ypf1'
        assert association.ymd_str(r1.date, "") == '20150305'

    for r in results:
        #print(str(r))
        sid = r.subject.id
        prov = r.provided_by
        assert prov == 'PomBase' or prov == 'UniProt'
        assert r.object.id.namespace == "GO"
        assert sid.namespace == 'PomBase' or (not is_gaf and sid.namespace == 'PR')
        if is_gaf:
            assert str(r.subject.taxon) =='NCBITaxon:4896'

    # for m in p.report.messages:
    #     print("MESSAGE: {}".format(m))
    print("MESSAGES (sample): {}".format(p.report.messages[0:5]))
    # Messages that aren't about upgrading qualifiers in rule 59 should be 0
    assert len([msg for msg in p.report.messages if msg["rule"] != 59 and msg["rule"] != 27]) == 0
    # print(p.report.to_markdown())


def test_flag_invalid_id():
    ont = OntologyFactory().create(ONT)
    p = GafParser()
    p.config.ontology = ont
    p._validate_ontology_class_id("FAKE:1", assocparser.SplitLine("fake", [""]*17, taxon="foo"))
    assert len(p.report.messages) == 1


def test_no_flag_valid_id():
    ont = OntologyFactory().create(ONT)
    p = GafParser()
    p.config.ontology = ont
    p._validate_ontology_class_id("GO:0000785", assocparser.SplitLine("fake", [""]*17, taxon="foo"))
    assert len(p.report.messages) == 0


def test_convert_gaf_to_gpad():
    p = GafParser()
    p.config.ecomap = EcoMap()
    w = GpadWriter()
    p2 = GpadParser()
    convert(POMBASE, p, w, p2)


def convert(file, p, w, p2):
    assocs = p.parse(file, skipheader=True)
    outfile = tempfile.NamedTemporaryFile(mode='w', delete=False)
    w.file = outfile
    for a in assocs:
        w.write_assoc(a)
    outfile.close()
    assocs2 = p2.parse(outfile.name)
    for a in assocs2:
        print("REPARSED: {}".format(a))
    assert len(assocs) == len(assocs2)


def test_invalid_goid_in_gpad():
    # Note: this ontology is a subset of GO extracted using the GAF, not GPAD
    p = GpadParser()
    p.config.ontology = OntologyFactory().create(ONT)
    results = p.parse(open(POMBASE_GPAD, "r"), skipheader=True)

    # we expect errors since ONT is not tuned for the GPAD file
    # for m in p.report.messages:
    #     print("MESSAGE: {}".format(m))
    assert len(p.report.messages) > 500
    print(p.report.to_markdown())


def test_validate_go_idspaces():
    ont = OntologyFactory().create(ONT)
    p = GafParser()
    p.config.class_idspaces = ['FOOZ']
    assocs = p.parse(open(POMBASE, "r"), skipheader=True)
    for m in p.report.messages:
        print("MESSAGE: {}".format(m))
    assert len(assocs) == 0
    assert len(p.report.messages) > 1
    summary = p.report.to_report_json()
    assert summary['associations'] == 0
    assert summary['lines'] > 300
    print(p.report.to_markdown())

    # ensure config is not preserved
    p = GafParser()
    assert p.config.class_idspaces is None


def test_qualifiers_gaf():
    p = GafParser()
    assocs = p.parse(open(QGAF, "r"), skipheader=True)
    neg_assocs = [a for a in assocs if a.negated == True]
    assert len(neg_assocs) == 3
    for a in assocs:
        print('REL: {}'.format(str(a.relation)))

    assert len([a for a in assocs if str(a.relation) == 'RO:0002326']) == 1

    # For the space in `colocalizes with`
    assert len(list(filter(lambda e: e["obj"] == "colocalizes with", p.report.to_report_json()["messages"]["gorule-0000001"]))) == 1
    assert len(list(filter(lambda e: e["obj"] == "involved_in", p.report.to_report_json()["messages"]["gorule-0000001"]))) == 1


def test_gaf_2_1_unknown_qualifier():
    line = ["UniProtKB", "P0AFI2", "parC", "Contributes_to", "GO:0003916", "PMID:1334483", "IDA", "", "F", "", "", "gene", "taxon:83333", "20081208", "EcoliWiki"]
    parsed = gafparser.to_association(line)

    assert parsed.skipped is True
    assert parsed.report.to_report_json()["messages"]["gorule-0000001"][0]["type"] == parsed.report.INVALID_QUALIFIER


def test_qualifiers_gaf_2_2():

    p = GafParser()

    assocs = p.parse(open("tests/resources/test-qualifiers-2.2.gaf"), skipheader=True)
    # NOT by itself is not allowed
    assert len(list(filter(lambda e: e["obj"] == "NOT", p.report.to_report_json()["messages"]["gorule-0000001"]))) == 1
    assert len(list(filter(lambda e: e["obj"] == "contributes_to|enables", p.report.to_report_json()["messages"]["gorule-0000001"]))) == 1
    assert len([a for a in assocs if association.Curie.from_str("RO:0004035") in a.qualifiers]) == 1


def test_gaf_2_1_creates_cell_component_closure():
    ontology = OntologyFactory().create("tests/resources/goslim_generic.json")
    closure = gafparser.protein_complex_sublcass_closure(ontology)
    # "GO:1902494" as an example that should be in the set
    assert "GO:0005840" in closure

    p = GafParser(config=assocparser.AssocParserConfig(ontology=ontology))
    with open("tests/resources/pombase_single.gaf") as gaf:
        # First line will be version declaration, triggering closure computation
        p.parse_line(gaf.readline())

    assert "GO:0005840" in p.cell_component_descendants_closure


def test_gaf_2_1_qualifiers_upconvert():
    line = ["SGD", "S000000819", "AFG3", "", "GO:0005840", "PMID:8681382|SGD_REF:S000055187", "IMP", "", "P", "Mitochondrial inner membrane m-AAA protease component", "YER017C|AAA family ATPase AFG3|YTA10", "gene", "taxon:559292", "20170428", "SGD"]
    parsed = gafparser.to_association(line)
    assoc = parsed.associations[0]

    ontology = OntologyFactory().create("tests/resources/goslim_generic.json")
    p = GafParser(config=assocparser.AssocParserConfig(ontology=ontology))
    p.make_internal_cell_component_closure()

    assoc = p.upgrade_empty_qualifier(assoc)
    assert assoc.qualifiers[0] == association.Curie(namespace="BFO", identity="0000050")


def test_gaf_2_1_upconvert_in_parse():
    gaf = io.StringIO("!gaf-version: 2.1\nSGD\tS000000819\tAFG3\t\tGO:0005840\tPMID:8681382|SGD_REF:S000055187\tIMP\t\tP\tMitochondrial inner membrane m-AAA protease component\tYER017C|AAA family ATPase AFG3|YTA10\tgene\ttaxon:559292\t20170428\tSGD")
    ontology = OntologyFactory().create("tests/resources/goslim_generic.json")
    p = GafParser(config=assocparser.AssocParserConfig(ontology=ontology))

    # We're 2.1, qualifier blank, cell component term from above, ontology defined: should upgrade
    assocs = p.parse(gaf, skipheader=True)
    assert assocs[0].relation == association.Curie(namespace="BFO", identity="0000050")


def test_gaf_2_1_simple_terms():
    line = ["SGD", "S000000819", "AFG3", "", "GO:0006259", "PMID:8681382|SGD_REF:S000055187", "IMP", "", "P", "Mitochondrial inner membrane m-AAA protease component", "YER017C|AAA family ATPase AFG3|YTA10", "gene", "taxon:559292", "20170428", "SGD"]
    ontology = OntologyFactory().create("tests/resources/goslim_generic.json")
    p = GafParser(config=assocparser.AssocParserConfig(ontology=ontology))
    p.make_internal_cell_component_closure()

    parsed = gafparser.to_association(line)
    assoc = p.upgrade_empty_qualifier(parsed.associations[0])
    assert assoc.qualifiers[0] == association.Curie(namespace="RO", identity="0002264")

    line = ["SGD", "S000000819", "AFG3", "", "GO:0042393", "PMID:8681382|SGD_REF:S000055187", "IMP", "", "P",
            "Mitochondrial inner membrane m-AAA protease component", "YER017C|AAA family ATPase AFG3|YTA10", "gene",
            "taxon:559292", "20170428", "SGD"]
    ontology = OntologyFactory().create("tests/resources/goslim_generic.json")
    p = GafParser(config=assocparser.AssocParserConfig(ontology=ontology))
    p.make_internal_cell_component_closure()

    parsed = gafparser.to_association(line)
    assoc = p.upgrade_empty_qualifier(parsed.associations[0])
    assert assoc.qualifiers[0] == association.Curie(namespace="RO", identity="0002327")

    line = ["SGD", "S000000819", "AFG3", "", "GO:0005773", "PMID:8681382|SGD_REF:S000055187", "IMP", "", "P",
            "Mitochondrial inner membrane m-AAA protease component", "YER017C|AAA family ATPase AFG3|YTA10", "gene",
            "taxon:559292", "20170428", "SGD"]
    ontology = OntologyFactory().create("tests/resources/goslim_generic.json")
    p = GafParser(config=assocparser.AssocParserConfig(ontology=ontology))
    p.make_internal_cell_component_closure()

    parsed = gafparser.to_association(line)
    assoc = p.upgrade_empty_qualifier(parsed.associations[0])
    assert assoc.qualifiers[0] == association.Curie(namespace="RO", identity="0001025")


def test_upgrade_qualifiers_for_biological_process():
    line = ["SGD", "S000000819", "AFG3", "", "GO:0008150", "PMID:8681382|SGD_REF:S000055187", "IMP", "", "P",
            "Mitochondrial inner membrane m-AAA protease component", "YER017C|AAA family ATPase AFG3|YTA10", "gene",
            "taxon:559292", "20170428", "SGD"]
    ontology = OntologyFactory().create("tests/resources/goslim_generic.json")
    p = GafParser(config=assocparser.AssocParserConfig(ontology=ontology))
    p.make_internal_cell_component_closure()

    parsed = gafparser.to_association(line)
    assoc = p.upgrade_empty_qualifier(parsed.associations[0])
    assert assoc.qualifiers[0] == association.Curie(namespace="RO", identity="0002331")


def test_upgrade_qualifiers_for_cell_component():
    line = ["SGD", "S000000819", "AFG3", "", "GO:0008372", "PMID:8681382|SGD_REF:S000055187", "IMP", "", "P",
            "Mitochondrial inner membrane m-AAA protease component", "YER017C|AAA family ATPase AFG3|YTA10", "gene",
            "taxon:559292", "20170428", "SGD"]
    ontology = OntologyFactory().create("tests/resources/goslim_generic.json")
    p = GafParser(config=assocparser.AssocParserConfig(ontology=ontology))
    p.make_internal_cell_component_closure()

    parsed = gafparser.to_association(line)
    assoc = p.upgrade_empty_qualifier(parsed.associations[0])
    assert assoc.qualifiers[0] == association.Curie(namespace="RO", identity="0002432")


def test_default_gaf_version():
    p = GafParser()
    assocs = p.parse(open("tests/resources/test-qualifiers-no-version.gaf"), skipheader=True)
    assert p.version == "2.1"


def parse_with2(f, p):
    ont = OntologyFactory().create(ONT)

    p.config.ontology = ont
    assocs = p.parse(open(f, "r"), skipheader=True)
    neg_assocs = [a for a in assocs if a.negated == True]
    assert len(neg_assocs) == 3
    for a in assocs:
        print('REL: {}'.format(a.relation))
    assert len([a for a in assocs if str(a.relation) == relations.lookup_label('involved_in')]) == 1
    assert len([a for a in assocs if str(a.relation) == relations.lookup_label('contributes_to')]) == 1


def test_errors_gaf():
    config = assocparser.AssocParserConfig(
        ecomap=EcoMap()
    )
    p = GafParser(config=config)
    assocs = p.parse(open("tests/resources/errors.gaf", "r"), skipheader=True)
    msgs = p.report.messages
    print(json.dumps(p.report.to_report_json(), indent=4))
    # print("MESSAGES: {}".format(len(msgs)))
    n_invalid_idspace = 0
    for m in msgs:
        print("MESSAGE: {}".format(m))
        if m['type'] == assocparser.Report.INVALID_IDSPACE:
            n_invalid_idspace += 1
    assert len(msgs) == 13
    assert n_invalid_idspace == 1
    assert len(assocs) == 1

    w = GafWriter()
    w.write(assocs)
    for a in assocs:
        if a.object_extensions != []:
            # our test file has no ORs, so in DNF this is always the first
            xs = a.object_extensions[0].elements
            print(xs)
            for x in xs:

                print('X: {}'.format(x))
                # ensure that invalid expressions have been eliminated
                assert x.relation == association.Curie("BFO", "0000050")
                assert x.term == association.Curie.from_str('X:1')
            assert len(xs) == 1


def test_alt_id_repair():
    p = GafParser()
    ont = OntologyFactory().create(ALT_ID_ONT)
    p.config.ecomap = EcoMap()
    p.config.ontology = ont

    gaf = io.StringIO("SGD\tS000000819\tAFG3\t\tGO:1\tPMID:8681382|SGD_REF:S000055187\tIMP\t\tP\tMitochondrial inner membrane m-AAA protease component\tYER017C|AAA family ATPase AFG3|YTA10\tgene\ttaxon:559292\t20170428\tSGD")
    assocs = p.parse(gaf, skipheader=True)
    # GO:1 is obsolete, and has replaced by GO:0034622, so we should see that class ID.
    assert assocs[0].object.id == association.Curie.from_str("GO:2")

    gaf = io.StringIO("SGD\tS000000819\tAFG3\t\tGO:4\tPMID:8681382|SGD_REF:S000055187\tIMP\t\tP\tMitochondrial inner membrane m-AAA protease component\tYER017C|AAA family ATPase AFG3|YTA10\tgene\ttaxon:559292\t20170428\tSGD")
    assocs = p.parse(gaf, skipheader=True)
    # GO:4 is obsolete due to it being merged into GO:3
    assert assocs[0].object.id == association.Curie.from_str("GO:3")


def test_gorule_repair():
    config = assocparser.AssocParserConfig(
        ontology=OntologyFactory().create("tests/resources/goslim_generic.json"),
        rule_set=assocparser.RuleSet.ALL
    )
    p = GafParser(config=config)
    # Here this gaf line has the wrong aspect, and should be picked up by gorule 28
    gaf = io.StringIO("PomBase\tSPCC962.06c\tbpb1\t\tGO:0005634\tPMID:20970342\tIPI\t\tP\tKH and CC/hC domain splicing factor Bpb1\tsf1|ods3\tprotein\ttaxon:4896\t20110804\tPomBase\tpart_of(GO:0007137)")
    assocs = p.parse(gaf, skipheader=True)

    assert assocs[0].aspect == "C"
    assert len(p.report.to_report_json()["messages"]["gorule-0000028"]) == 1
    assert p.report.to_report_json()["messages"]["gorule-0000028"][0]["type"] == assocparser.Report.VIOLATES_GO_RULE


def test_bad_date():
    p = GafParser()
    assoc_result = p.parse_line("PomBase\tSPAC25B8.17\typf1\t\tGO:0000007\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:4896\tTODAY\tPomBase\tfoo(X:1)")
    assert assoc_result.skipped == True
    assert assoc_result.associations == []
    
def test_bad_gene_symbol():
    p = GafParser()
    assoc_result = p.parse_line("PomBase\tSPAC25B8.17\ta|pipeisnotallowed\t\tGO:0000007\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:4896\t20231110\tPomBase\tfoo(X:1)")
    assert assoc_result.skipped == True
    assert assoc_result.associations == []     
    
def test_bad_go_id():
    p = GafParser()
    assoc_result = p.parse_line("PomBase\tSPAC25B8.17\typf1\t\tINVALID:0000007\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:4896\t20231110\tPomBase\tfoo(X:1)")
    assert assoc_result.skipped == True
    assert assoc_result.associations == []    

def test_bad_taxon():
    p = GafParser()
    assoc_result = p.parse_line("PomBase\tSPAC25B8.17\typf1\t\tGO:0000007\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:1abc\t20231003\tPomBase\tfoo(X:1)")
    assert assoc_result.skipped == True
    assert assoc_result.associations == []

def test_subject_extensions():
    p = GafParser()
    assoc_result = p.parse_line("PomBase\tSPAC25B8.17\typf1\t\tGO:0000007\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:4896\t20181024\tPomBase\tpart_of(X:1)\tUniProtKB:P12345")
    assert len(assoc_result.associations[0].subject_extensions) == 1

    subject_extensions = assoc_result.associations[0].subject_extensions
    gene_product_form_id = subject_extensions[0].term
    assert gene_product_form_id == association.Curie.from_str("UniProtKB:P12345")


def test_bad_withfrom():
    p = GafParser()
    # With/from has no identity portion after the namespace
    assoc_result = p.parse_line("PomBase\tSPAC25B8.17\typf1\t\tGO:0000007\tGO_REF:0000024\tISO\tSGD:\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:4896\t20181024\tPomBase")
    assert assoc_result.associations == []
    assert p.report.to_report_json()["messages"]["gorule-0000001"][0]["obj"] == "SGD:"


def test_obsolete_replair_of_withfrom():
    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(OBSOLETE_ONT)))
    assocs = p.parse(open(ZFIN_GAF, "r"), skipheader=True)
    assert assocs[0].evidence.with_support_from == [ConjunctiveSet(elements=[Curie(namespace='GO', identity='0005912')])]

    # Reset parser report
    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(OBSOLETE_ONT)))
    p.version = "2.2"
    obsolete_no_replacement_line = "FB\tFBgn0003334\tScm\tlocated_in\tGO:0005634\tFB:FBrf0179383|PMID:15280237\tIC\tGO:0016458\tC\tSex comb on midleg\tCG9495|SCM|Sex Comb on Midleg|Sex Comb on the Midleg|Sex combs on midleg|Sex combs on midlegs|Su(z)302|l(3)85Ef|scm|sex comb on midleg\tprotein\ttaxon:7227\t20050203\tUniProt\t\t"
    assoc_result = p.parse_line(obsolete_no_replacement_line)
    assert assoc_result.associations == []
    assert p.report.to_report_json()["messages"]["gorule-0000020"][0]["obj"] == "GO:0016458"
    
    
def test_invalid_db_type():
    #gene_product gets mapped to gene_product
    line = ["UniProtKB", "P0AFI2", "parC", "", "GO:0003916", "PMID:1334483", "IDA", "", "F", "", "", "gene_product", "taxon:83333", "20081208", "EcoliWiki"]
    parsed = gafparser.to_association(line)
    assoc = parsed.associations[0]
    assert assoc.subject.type == [association.map_gp_type_label_to_curie('gene_product')]

    #protein gets mapped to protein
    line = ["UniProtKB", "P0AFI2", "parC", "", "GO:0003916", "PMID:1334483", "IDA", "", "F", "", "", "protein", "taxon:83333", "20081208", "EcoliWiki"]
    parsed = gafparser.to_association(line)
    assoc = parsed.associations[0]
    assert assoc.subject.type == [association.map_gp_type_label_to_curie('protein')]
    
    #Unhandled types get mapped to 'gene_product'
    line = ["UniProtKB", "P0AFI2", "parC", "", "GO:0003916", "PMID:1334483", "IDA", "", "F", "", "", "invalid_gene_product", "taxon:83333", "20081208", "EcoliWiki"]
    parsed = gafparser.to_association(line)
    assoc = parsed.associations[0]
    assert assoc.subject.type == [association.map_gp_type_label_to_curie('gene_product')]
    assert parsed.report.to_report_json()["messages"]["gorule-0000001"][0]["type"] == parsed.report.INVALID_SUBJECT_TYPE
    
    #'lnc_RNA' gets repaired to 'lncRNA'
    line = ["UniProtKB", "P0AFI2", "parC", "", "GO:0003916", "PMID:1334483", "IDA", "", "F", "", "", "lnc_RNA", "taxon:83333", "20081208", "EcoliWiki"]
    parsed = gafparser.to_association(line)
    assoc = parsed.associations[0]
    assert assoc.subject.type == [association.map_gp_type_label_to_curie('lncRNA')]                
    assert parsed.report.to_report_json()["messages"]["gorule-0000001"][0]["type"] == parsed.report.INVALID_SUBJECT_TYPE
    
 

def test_subject_extensions_bad_curie():
    """
    Offending field is `GDP_bound`
    """
    p = GafParser()
    assoc_result = p.parse_line("PomBase\tSPBC1289.03c\tspi1\t\tGO:0005515\tPMID:18422602\tIPI\tPomBase:SPAC25A8.01c\tF\tRan GTPase Spi1\t\tprotein\ttaxon:4896\t20080718\tPomBase\t\tGDP_bound")
    assert assoc_result.associations == []
    assert assoc_result.skipped == True
    assert len(p.report.to_report_json()["messages"]["gorule-0000001"]) == 1
    assert p.report.to_report_json()["messages"]["gorule-0000001"][0]["type"] == p.report.INVALID_ID
    assert p.report.to_report_json()["messages"]["gorule-0000001"][0]["obj"] == "GDP_bound"
    print(json.dumps(p.report.to_report_json(), indent=4))


def test_object_extensions():
    p = GafParser()
    assoc_result = p.parse_line("PomBase\tSPAC25B8.17\typf1\t\tGO:0000007\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:4896\t20181024\tPomBase\tpart_of(X:1)\tUniProtKB:P12345")
    print(p.report.to_markdown())
    assert len(assoc_result.associations[0].object_extensions) > 0
    object_extensions = [
        association.ConjunctiveSet([
            association.ExtensionUnit(association.Curie("BFO", "0000050"), association.Curie("X", "1"))
        ])
    ]
    assert assoc_result.associations[0].object_extensions == object_extensions


def test_object_extensions_error():
    p = GafParser()
    assoc_result = p.parse_line("PomBase\tSPAC25B8.17\typf1\t\tGO:0000007\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:4896\t20181024\tPomBase\tpart_of(X)\tUniProtKB:P12345")
    assert len(p.report.to_report_json()["messages"]["gorule-0000001"]) == 1


def test_obsolete_replair_of_extensions():
    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(OBSOLETE_ONT))) 
    assoc_result = p.parse_line("ZFIN\tZDB-GENE-980526-362\tctnnb1\t\tGO:0007155\tPMID:15494018\tIC\tGO:0005913\tP\tcatenin (cadherin-associated protein), beta 1\tctnnb|id:ibd2058|wu:fb73e10|wu:fi81c06|wu:fk25h01\tprotein_coding_gene\ttaxon:7955\t20041026\tZFIN\tpart_of(GO:0005913)\tUniProtKB:P12345")
    print(p.report.to_markdown())
    assert len(assoc_result.associations[0].object_extensions) > 0
    object_extensions = [
        association.ConjunctiveSet([
            association.ExtensionUnit(association.Curie("BFO", "0000050"), association.Curie("GO", "0005912"))
        ])
    ]
    assert assoc_result.associations[0].object_extensions == object_extensions
    
    #Reset the report
    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(OBSOLETE_ONT))) 
    assoc_result = p.parse_line("ZFIN\tZDB-GENE-980526-362\tctnnb1\t\tGO:0007155\tPMID:15494018\tIC\tGO:0005912\tP\tcatenin (cadherin-associated protein), beta 1\tctnnb|id:ibd2058|wu:fb73e10|wu:fi81c06|wu:fk25h01\tprotein_coding_gene\ttaxon:7955\t20041026\tZFIN\tpart_of(GO:0005913)|part_of(GO:0016458)\tUniProtKB:P12345")
    print(p.report.to_markdown())
    assert assoc_result.associations == []
    assert p.report.to_report_json()["messages"]["gorule-0000020"][0]["obj"] == "GO:0016458"    
    assert p.report.to_report_json()["messages"]["gorule-0000020"][1]["obj"] == "GO:0005913"
    
    #Reset the report
    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(OBSOLETE_ONT))) 
    assoc_result = p.parse_line("ZFIN\tZDB-GENE-980526-362\tctnnb1\t\tGO:0007155\tPMID:15494018\tIC\tGO:0005912\tP\tcatenin (cadherin-associated protein), beta 1\tctnnb|id:ibd2058|wu:fb73e10|wu:fi81c06|wu:fk25h01\tprotein_coding_gene\ttaxon:7955\t20041026\tZFIN\tpart_of(GO:0005913)|part_of(GO:0007155),occurs_in(Y:1)\tUniProtKB:P12345")
    print(p.report.to_markdown())
    assert len(assoc_result.associations[0].object_extensions) > 0
    object_extensions = [
        association.ConjunctiveSet([association.ExtensionUnit(association.Curie("BFO", "0000050"), association.Curie("GO", "0005912"))]),
        association.ConjunctiveSet([association.ExtensionUnit(association.Curie("BFO", "0000050"), association.Curie("GO", "0007155")),association.ExtensionUnit(association.Curie("BFO", "0000066"), association.Curie("Y", "1"))])        
    ]
    assert assoc_result.associations[0].object_extensions == object_extensions    
    
    #Reset the report
    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(OBSOLETE_ONT)))
    assoc_result = p.parse_line("ZFIN\tZDB-GENE-980526-362\tctnnb1\t\tGO:0007155\tPMID:15494018\tIC\tGO:0005912\tP\tcatenin (cadherin-associated protein), beta 1\tctnnb|id:ibd2058|wu:fb73e10|wu:fi81c06|wu:fk25h01\tprotein_coding_gene\ttaxon:7955\t20041026\tZFIN\tpart_of(GO:0016458)\tUniProtKB:P12345")
    print(p.report.to_markdown())
    assert assoc_result.associations == []
    assert p.report.to_report_json()["messages"]["gorule-0000020"][0]["obj"] == "GO:0016458"    
    

def test_factory():
    afa = AssociationSetFactory()
    ont = OntologyFactory().create(ONT)
    aset = afa.create_from_file(POMBASE, ontology=ont, skim=False)

    found = 0
    for s in aset.subjects:
        print('{} {}'.format(s, aset.label(s)))
        for c in aset.annotations(s):
            print('  {} {}'.format(c, ont.label(c)))
            for a in aset.associations(s, c):
                e = a['evidence']
                print('    {} {} {}'.format(e['type'], e['with_support_from'], e['has_supporting_reference']))
                if s == 'PomBase:SPBC2D10.10c' and c == 'GO:0005730':
                    if e['type'] == 'ISO':
                        if e['with_support_from'] == ['SGD:S000002172'] and e['has_supporting_reference'] == ['GO_REF:0000024']:
                            found +=1
                            logging.info('** FOUND: {}'.format(a))
                    if e['type'] == 'IDA':
                        if e['has_supporting_reference'] == ['PMID:16823372']:
                            found +=1
                            logging.info('** FOUND: {}'.format(a))

    assert len(aset.associations_by_subj) > 0
    assert found == 2
    
def test_id_syntax():
    database_id_syntax_lookups = {}
    go_types = {}
    pattern = '\\d{7}'
    go_types['molecular_function'] = re.compile(pattern)
    go_types['biological_process'] = re.compile(pattern)
    go_types['cellular_component'] = re.compile(pattern)      
    database_id_syntax_lookups['GO'] = go_types
    
    pmid_types = {}
    pmid_types['entity'] = re.compile('[0-9]+')
    database_id_syntax_lookups['PMID'] = pmid_types
    
    pombase_types = {}
    pombase_types['entity'] = re.compile('S\\w+(\\.)?\\w+(\\.)?')
    database_id_syntax_lookups['PomBase'] = pombase_types    
    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(ONT), db_type_name_regex_id_syntax=database_id_syntax_lookups))

    assoc_result = p.parse_line("PomBase\tSPBC1289.03c\tspi1\t\tGO:0005515\tPMID:18422602\tIPI\tPomBase:SPAC25A8.01c\tF\tRan GTPase Spi1\t\tprotein\ttaxon:4896\t20080718\tPomBase\t")
    assert len(assoc_result.associations) == 1
    assert assoc_result.skipped == False
    messages = p.report.to_report_json()["messages"]
    assert "gorule-0000027" not in messages

    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(ONT), db_type_name_regex_id_syntax=database_id_syntax_lookups))
    assoc_result = p.parse_line("PomBase\tSPBC1289.03c\tspi1\t\tGO:0005515\tPMID:PMID:18422602\tIPI\tPomBase:SPAC25A8.01c\tF\tRan GTPase Spi1\t\tprotein\ttaxon:4896\t20080718\tPomBase\t")
    assert len(assoc_result.associations) == 1
    assert assoc_result.skipped == False
    messages = p.report.to_report_json()["messages"]
    assert len(messages["gorule-0000027"]) == 1
    assert messages["gorule-0000027"][0]["obj"] == "PMID:PMID:18422602" 
    
    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(ONT), db_type_name_regex_id_syntax=database_id_syntax_lookups))
    assoc_result = p.parse_line("PomBase\tSPBC1289.03c\tspi1\t\tGO:0005515\tPMID:0.\tIPI\tPomBase:SPAC25A8.01c\tF\tRan GTPase Spi1\t\tprotein\ttaxon:4896\t20080718\tPomBase\t")
    assert len(assoc_result.associations) == 1
    assert assoc_result.skipped == False
    messages = p.report.to_report_json()["messages"]
    assert len(messages["gorule-0000027"]) == 1
    assert messages["gorule-0000027"][0]["obj"] == "PMID:0."

    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(ONT), db_type_name_regex_id_syntax=database_id_syntax_lookups))
    assoc_result = p.parse_line("PomBase\tSPBC1289.03c\tspi1\t\tGO:0005515\tPMID:x18422602\tIPI\tPomBase:SPAC25A8.01c\tF\tRan GTPase Spi1\t\tprotein\ttaxon:4896\t20080718\tPomBase\t")
    assert len(assoc_result.associations) == 1
    assert assoc_result.skipped == False
    messages = p.report.to_report_json()["messages"]
    assert len(messages["gorule-0000027"]) == 1
    assert messages["gorule-0000027"][0]["obj"] == "PMID:x18422602"

    p = GafParser(config=assocparser.AssocParserConfig(
        ontology=OntologyFactory().create(ONT), db_type_name_regex_id_syntax=database_id_syntax_lookups))
    assoc_result = p.parse_line("PomBase\tSPBC1289.03c\tspi1\t\tGO:0005515\tBLA:18422602\tIPI\tPomBase:SPAC25A8.01c\tF\tRan GTPase Spi1\t\tprotein\ttaxon:4896\t20080718\tPomBase\t")
    assert len(assoc_result.associations) == 1
    assert assoc_result.skipped == False
    messages = p.report.to_report_json()["messages"]    
    assert len(messages["gorule-0000027"]) == 1
    assert messages["gorule-0000027"][0]["obj"] == "BLA:18422602" 


def test_gaf_gpi_bridge():
    gaf = ["MGI", "MGI:1923503", "0610006L08Rik", "enables", "GO:0003674", "MGI:MGI:2156816|GO_REF:0000015", "ND", "",
           "F", "RIKEN cDNA 0610006L08 gene", "", "gene", "taxon:10090", "20120430", "MGI", "", ""]
    association = gafparser.to_association(gaf, qualifier_parser=assocparser.Qualifier2_2()).associations[0]
    bridge = gafgpibridge
    entity = bridge.convert_association(association)
    assert entity.get("type") == ["gene"]


if __name__ == "__main__":
    pytest.main(args=["tests/test_gafparser.py::test_parse_gaf"])



import json
from typing import Dict, NewType, List

from ontobio.model.association import GoAssociation, gp_type_label_to_curie


class Entity(dict):

    def __init__(self, d):
        super(Entity, self).__init__(d)

    def __hash__(self):
        d = json.dumps(self, sort_keys=True)
        return hash(d)


def convert_association(association):
    """
    'id' is already `join`ed in both the Association and the Entity,
    so we don't have to worry about what that looks like. We assume
    it's correct.
    """

    if isinstance(association, GoAssociation):
        # print(json.dumps(association, indent=4))
        gpi_obj = {
            'id': str(association.subject.id),
            'label': association.subject.label,  # db_object_symbol,
            'full_name': association.subject.fullname,  # db_object_name,
            'synonyms': association.subject.synonyms,
            'type': [gp_type_label_to_curie(association.subject.type[0])], #db_object_type,
            'parents': "",  # GAF does not have this field, but it's optional in GPI
            'xrefs': "",  # GAF does not have this field, but it's optional in GPI
            'taxon': {
                'id': str(association.subject.taxon)
            }
        }
        return Entity(gpi_obj)

    return None


class GafGpiBridge(object):

    def __init__(self):
        self.cache = []

    def entities(self):
        return list(self.cache)
